# -*- coding: utf-8 -*-
"""
oathldap_tool.ykadd -- sub-command for adding LDAP entry for Yubikey token
"""

#---------------------------------------------------------------------------
# Imports
#---------------------------------------------------------------------------

import time
import sys

# from python-usb package
from usb.core import USBError

# from ldap0 package
import ldap0
import ldap0.base
from ldap0 import LDAPError
from ldap0.ldapurl import LDAPUrl
from ldap0.ldif import LDIFParser
from ldap0.tmpl import TemplateEntry

# from oathldap_tool package
from .__about__ import __version__
from .yubikey import YubiKeySearchError, YKTokenDevice
from . import SEP_LINE, ErrorExit, cli_output
from . import interactive_ldapconnect

#---------------------------------------------------------------------------
# Constants
#---------------------------------------------------------------------------

OWNER_FILTER_TMPL = (
    '(&'
      '(objectClass=inetOrgPerson)'
      '(|'
        '(uid={0})'
        '(mail={0})'
        '(uniqueIdentifier={0})'
        '(employeeNumber={0})'
      ')'
    ')'
)

#---------------------------------------------------------------------------
# Functions
#---------------------------------------------------------------------------

def ykadd(command_name, args):
    """
    Adds Yubikey token entry to OATH-LDAP server
    """
    cli_output(SEP_LINE, lf_before=0, lf_after=0)
    cli_output('OATH-LDAP {0} v{1}'.format(command_name, __version__))
    cli_output(SEP_LINE, lf_before=0)

    try:

        # read the LDIF template
        try:
            with open(args.ldif_template, 'rb') as ldif_file:
                t_dn, t_entry = LDIFParser(ldif_file).list_entry_records()[0]
        except IOError as err:
            raise ErrorExit('Unable to read LDIF template: %s' % (err))

        # parse the LDAP URL
        ldap_url = LDAPUrl(args.ldap_url)

        # Open OATH-LDAP connection interactively asking for admin password
        oath_ldap = interactive_ldapconnect(
            ldap_url.connect_uri(),
            args.admin_dn,
        )

        owner_filter = OWNER_FILTER_TMPL.format(ldap0.filter.escape_str(args.person_id))
        if ldap_url.filterstr is not None:
            owner_filter = '(&{0}{1})'.format(ldap_url.filterstr, owner_filter)
        owner = oath_ldap.find_unique_entry(
            ldap_url.dn,
            ldap_url.scope or ldap0.SCOPE_SUBTREE,
            owner_filter,
            attrlist=['1.1'],
        )

        if args.serial_nr:
            yk_serial = int(args.serial_nr)
        else:
            # try to find a device and output info
            input('Remove Yubikey used for login, press [ENTER] to continue...')
            sys.stdout.write('\nWaiting for single Yubikey device.')
            while True:
                try:
                    yk_device = YKTokenDevice.search()
                except (YubiKeySearchError, USBError):
                    sys.stdout.write('.')
                    sys.stdout.flush()
                    time.sleep(0.8)
                else:
                    sys.stdout.write('\n')
                    break
            yk_serial = yk_device.key.serial()
            cli_output('Found Yubikey device no. %s' % yk_serial)
            cli_output(yk_device.info_msg())

        token_dn, token_entry = TemplateEntry(
            t_dn.decode('utf-8'),
            ldap0.base.decode_entry_dict(t_entry),
        ).ldap_entry(
            dict(
                search_base=ldap_url.dn,
                owner=owner.dn_s,
                serial=yk_serial,
            )
        )
        token_dn = ','.join((token_dn, ldap_url.dn))

        # finally add OATH-LDAP token entry
        oath_ldap.add_s(token_dn, token_entry)
        cli_output('Added OATH-LDAP entry %r' % (token_dn))
        oath_ldap.unbind_s()

    except USBError as err:
        cli_output(str(err))

    except ldap0.ALREADY_EXISTS:
        cli_output('Entry %r already exists!' % (token_dn))

    except LDAPError as err:
        cli_output(str(err))

    except ErrorExit as err:
        cli_output(str(err))

    except YubiKeySearchError as err:
        cli_output(str(err))

    # end of ykadd()
