# -*- coding: utf-8 -*-
"""
oathldap_tool.ykinit -- sub-command for initializing a Yubikey token
"""

#---------------------------------------------------------------------------
# Imports
#---------------------------------------------------------------------------

import time
import sys
#import signal
import os
import json

# from python-usb package
from usb.core import USBError

# from jwcrypto package
from jwcrypto.jwk import JWK
from jwcrypto.jwe import JWE

# from ldap0 package
import ldap0
from ldap0 import LDAPError
from ldap0.ldapurl import LDAPUrl
from ldap0.ldapobject import ReconnectLDAPObject
from ldap0.pw import random_string

# from oathldap package
from .__about__ import __version__
from .yubikey import YK_PASSWORD_ALPHABET, YubiKeySearchError, YKTokenDevice
from . import SEP_LINE, ErrorExit, cli_output


#---------------------------------------------------------------------------
# Constants
#---------------------------------------------------------------------------

JWE_DEFAULT_ALG = 'RSA-OAEP'
JWE_DEFAULT_ENC = 'A256GCM'

ABORT_MSG = 'Aborted -> please insert new Yubikey'

# string template for bind-DN of an oathToken entry
OATH_TOKEN_BINDDN_TMPL = r'serialNumber=yubikey-{unique_id},{search_base}'

#---------------------------------------------------------------------------
# Classes
#---------------------------------------------------------------------------

class OathLDAPConn(ReconnectLDAPObject):
    """
    OATH-LDAP connection
    """

    def __init__(self, ldap_uri, who, cred, **kwargs):
        ReconnectLDAPObject.__init__(self, ldap_uri, **kwargs)
        self.set_option(ldap0.OPT_NETWORK_TIMEOUT, 15.0)
        self.simple_bind_s(who, cred)
        self.bound_dn = self.whoami_s()[3:]
        # end of __init__()

    def get_hotp_params(self, token_dn=None):
        """
        read some HOTP parameters from token and policy entries
        """
        token = self.read_s(
            token_dn or self.bound_dn,
            filterstr='(objectClass=oathHOTPToken)',
            attrlist=[
                'oathHOTPParams',
                'oathTokenIdentifier',
            ],
        )
        token_params = self.read_s(
            token.entry_s['oathHOTPParams'][0],
            filterstr='(objectClass=oathHOTPParams)',
            attrlist=[
                'oathEncKey',
                'oathHMACAlgorithm',
                'oathOTPLength',
            ],
        )
        if 'oathEncKey' in token_params.entry_s:
            enc_key = JWK(**json.loads(token_params.entry_s['oathEncKey'][0]))
        else:
            enc_key = None
        return (
            token.entry_s.get('oathTokenIdentifier', [''])[0],
            token_params.entry_s['oathHMACAlgorithm'][0],
            int(token_params.entry_s['oathOTPLength'][0]),
            enc_key,
        )

    def update_token(self, otp_secret, token_pin, token_dn=None):
        """
        Write OATH-LDAP attributes, especially the already encrypted
        :otp_secret:, to the OATH-LDAP token entry this connection is bound as.
        """
        # Actually send LDAP modify request
        self.modify_s(
            token_dn or self.bound_dn,
            [
                (ldap0.MOD_ADD, b'oathHOTPCounter', [b'0']),
                (ldap0.MOD_REPLACE, b'oathTokenPIN', [token_pin.encode('utf-8')]),
                (ldap0.MOD_ADD, b'oathSecret', [otp_secret.encode('utf-8')]),
            ],
        )
        # end of update_token()

#---------------------------------------------------------------------------
# Functions
#---------------------------------------------------------------------------

def jwk_read(key_filename):
    """
    reads a JWK public key from file and returns the JWK object
    """
    with open(key_filename, 'rb') as pubkey_file:
        pk_json = pubkey_file.read()
    return JWK(**json.loads(pk_json))


def jwe_encrypt(jwk_obj, plaintext, alg=JWE_DEFAULT_ALG, enc=JWE_DEFAULT_ENC):
    """
    returns JWE strings with :plaintext: being asymmetrically encrypted
    with public key read from file with filename :key_filename:
    """
    jwe_obj = JWE(plaintext=plaintext)
    jwe_obj.add_recipient(
        jwk_obj,
        header=json.dumps(dict(
            alg=alg,
            enc=enc,
            kid=jwk_obj.key_id,
        )),
    )
    return jwe_obj.serialize() # end of jwe_encrypt()


def ask_access_code():
    """
    interactively ask for current slot password repeately in case of input error
    """
    current_access_code = input('Provide old slot password --> ')
    while current_access_code \
          and not (len(current_access_code) == 6 and current_access_code.isalnum()):
        current_access_code = input('Wrong format of slot password => try again')
    return current_access_code


def interactive_ldapconnect(ldap_url, serial):
    """
    returns LDAPObject connected as token specified by :serial: after
    interactively asking the user for the two enrollment password parts
    """
    ldap_url_obj = LDAPUrl(ldap_url)
    ldap_who = OATH_TOKEN_BINDDN_TMPL.format(
        unique_id=serial,
        search_base=ldap_url_obj.dn
    )
    while True:
        cli_output(
            (
                'Enter both enrollment password parts for Yubikey no. %s '
                'or remove YubiKey device and hit [ENTER] to rescan'
                '\nLDAP-DN: %r'
            ) % (serial, ldap_who)
        )
        ldap_initpw1 = input("Part #1 --> ")
        if not ldap_initpw1:
            raise ErrorExit(ABORT_MSG)
        ldap_initpw2 = input("Part #2 --> ")
        if not ldap_initpw2:
            raise ErrorExit(ABORT_MSG)
        try:
            oath_ldap = OathLDAPConn(
                ldap_url_obj.connect_uri(),
                ldap_who,
                ldap_initpw1+ldap_initpw2,
            )
        except ldap0.INVALID_CREDENTIALS:
            cli_output('Enrollment password(s) wrong => try again')
        else:
            break
    return oath_ldap # end of interactive_ldapconnect()


def ykinit_once(args, jwk_obj):
    """
    Run as stand-alone script
    """

    try:

        # try to find a device and output info
        sys.stdout.write('\nWaiting for single Yubikey device.')
        while True:
            try:
                yk_device = YKTokenDevice.search()
            except (YubiKeySearchError, USBError):
                sys.stdout.write('.')
                sys.stdout.flush()
                time.sleep(0.8)
            else:
                sys.stdout.write('\n')
                break

        access_code = args.access_code or random_string(YK_PASSWORD_ALPHABET, 6)
        if len(access_code) != 6 or not access_code.isalnum():
            raise ErrorExit(
                'Slot password must be exactly six (6) alpha-numeric ASCII characters!'
            )

        cli_output('Found Yubikey device no. %s' % yk_device.key.serial())
        cli_output(yk_device.info_msg())

        # if there are enabled slots we need the currently valid device password
        enabled_slots = yk_device.enabled_slots()
        if enabled_slots:
            confirm_input = input(
                'Confirm with word "reset" to clear the device '
                'or remove Yubikey device and hit [ENTER] to rescan --> '
            )
            if confirm_input.lower() != 'reset':
                raise ErrorExit(ABORT_MSG)
            if args.current_access_code:
                current_access_code = args.current_access_code
            else:
                current_access_code = ask_access_code()
            # clear enabled slots here
            cli_output('Clearing slot(s) %s...' % (enabled_slots))
            for slot in yk_device.enabled_slots():
                yk_device.reset_slot(slot, current_access_code)
            yk_device.clear(current_access_code)

        # save for later check to prevent user from accidently changing device
        yk_serial = yk_device.key.serial()

        del yk_device

        clear_time = time.time()

        # Open OATH-LDAP connection interactively asking for
        # enrollment password for this particular device
        oath_ldap = interactive_ldapconnect(args.ldap_url, yk_serial)

        # get some init parameters from OATH-LDAP
        oath_identifier, _, otp_length, oath_enckey = oath_ldap.get_hotp_params()
        if oath_enckey is None:
            oath_enckey = jwk_obj

        if oath_enckey is None:
            raise ErrorExit('No public key available to encrypt shared secret!')

        cli_output('Key used for encrypting secrets:')
        cli_output(
            'kid: %s\nthumbprint: %s' % (oath_enckey.key_id, oath_enckey.thumbprint()),
            lf_before=0,
        )

        # generate new shared secret (OTP seed)
        oath_secret = os.urandom(20)

        # write new encrypted shared secret and
        # new encrypted device password into OATH-LDAP token entry
        oath_ldap.update_token(
            jwe_encrypt(oath_enckey, oath_secret),
            jwe_encrypt(oath_enckey, access_code),
        )
        cli_output('Updated secret in OATH-LDAP entry %r' % (oath_ldap.bound_dn))
        oath_ldap.unbind_s()

        # wait some for the cleared device time if still needed
        sleep_time = 5.0 - (time.time() - clear_time)
        if sleep_time >= 0.0:
            cli_output('Waiting %0.1f secs until Yubikey reconnect...' % (sleep_time))
            time.sleep(sleep_time)

        # Search the device again
        yk_device = YKTokenDevice.search()
        if yk_serial != yk_device.key.serial():
            # better safe than sorry!
            raise ErrorExit('Device serial no. changed!')

        # now finally initialize the device
        cli_output('Initialize Yubikey device no. %s' % yk_device.key.serial())
        cli_output(
            '*** Write down new password for Yubikey %s and keep in safe place -> %s ***' % (
                yk_device.key.serial(),
                access_code,
            ),
            lf_before=2,
            lf_after=2,
        )
        yk_device.initialize(
            oath_secret,
            otp_length,
            oath_identifier,
            access_code,
        )

    except USBError as err:
        cli_output(str(err))

    except LDAPError as err:
        cli_output(str(err))

    except ErrorExit as err:
        cli_output(str(err))

    except YubiKeySearchError as err:
        cli_output(str(err))

    else:
        cli_output('Yubikey successfully initialized. Unplug it right now!', lf_after=2)
        if args.continous:
            input('Hit a key to initialize another Yubikey token...')

    cli_output(SEP_LINE)

    # end of ykinit_once()


def ykinit(command_name, args):
    """
    Initializes a Yubikey token for HOTP and sends encrypted shared secret
    to OATH-LDAP server after authenticating as the token entry with the
    enrollment password
    """
#    signal.signal(signal.SIGINT, signal.SIG_IGN)
#    signal.signal(signal.SIGTSTP, signal.SIG_IGN)
    cli_output(SEP_LINE, lf_before=0, lf_after=0)
    cli_output('OATH-LDAP {0} v{1}'.format(command_name, __version__))
    cli_output(SEP_LINE, lf_before=0)

    if args.pubkey:
        cli_output('Read JWK public key file %s' % (args.pubkey))
        try:
            jwk_obj = jwk_read(args.pubkey)
        except IOError as err:
            cli_output('Error reading JWK public key file %s -> %r' % (args.pubkey, err))
            sys.exit(1)
        cli_output('kid: %s\nthumbprint: %s' % (jwk_obj.key_id, jwk_obj.thumbprint()), lf_before=0)
    else:
        cli_output('No JWK public key specified')
        jwk_obj = None

    while True:
        ykinit_once(args, jwk_obj)
        if not args.continous:
            break

    # end of ykinit()
