# -*- coding: utf-8 -*-
"""
oathldap_tool.yubikey -- low-level Yubikey stuff
"""

import time
from binascii import unhexlify

# from python-yubico package
import yubico.yubikey_defs
from yubico.yubikey_base import YubiKeyError
from yubico import find_yubikey
from yubico.yubikey_config import YubiKeyConfig
from yubico.yubikey_neo_usb_hid import YubiKeyNEO_NDEF, YubiKeyNEO_DEVICE_CONFIG


__all__ = [
    'YK_SCAN_CODE',
    'YK_PASSWORD_ALPHABET',
    'YubiKeySearchError',
    'YKTokenDevice',
]

YK_SCAN_CODE = 'h:9284978b2c869291898c8a2c9a849691972c868f88849588871b2c87922c8c972c848a848c91'

YK_PASSWORD_ALPHABET = 'abcdefghijkmnpqrstuvwxyz23456789'


class YubiKeySearchError(Exception):
    """
    to be raised if search for a single Yubikey device failed
    """


class YKTokenDevice:
    """
    a single slot of a Yubikey token
    """
    find_yubikey_max_skip = 6

    def __init__(self, key):
        self.key = key

    @classmethod
    def search(cls):
        """
        init instance by searching for exactly one device
        """
        yk_devices = []
        skip = 0
        while True:
            try:
                yk_dev = find_yubikey(skip=skip)
            except YubiKeyError as yk_error:
                if yk_error.reason != 'No YubiKey found':
                    raise yk_error
                # check whether to proceed with search
                if skip >= cls.find_yubikey_max_skip:
                    break
            else:
                yk_devices.append(yk_dev)
            skip += 1
        if not yk_devices:
            raise YubiKeySearchError('No Yubikey devices found!')
        if len(yk_devices) > 1:
            raise YubiKeySearchError('More than one Yubikey found. Only one allowed!')
        return cls(yk_devices[0])

    def info(self):
        """
        return dict with various device information
        """
        return dict(
            model=self.key.model,
            serial=self.key.serial(),
            version=self.key.version(),
            status=self.key.status(),
        )

    def info_msg(self):
        """
        returns string with info about the device
        """
        info = self.info()
        info.update(
            slots=', '.join(map(str, info['status'].valid_configs())) or 'none',
        )
        return '{model} no. {serial} {version} // Enabled slots: {slots}'.format(**info)

    def enabled_slots(self):
        """
        returns list of numbers of enabled/configured slots
        """
        return self.key.status().valid_configs()

    def _write_cfg(self, ykcfg, slot):
        """
        write Yubikey configuration given in :cfg: to :slot:
        """
        if isinstance(ykcfg, YubiKeyConfig):
            self.key.write_config(ykcfg, slot)
        elif isinstance(ykcfg, YubiKeyNEO_NDEF):
            self.key.write_ndef(ykcfg, slot)
        elif isinstance(ykcfg, YubiKeyNEO_DEVICE_CONFIG):
            self.key.write_device_config(ykcfg)
        else:
            raise Exception('Wrong config type: %s' % (repr(ykcfg)))
        # end of _write_cfg()

    def init_hotp(
            self,
            slot,
            oath_secret,
            otp_len,
            oath_tokenid,
            access_code,
        ):
        """
        Initialize a device slot for HOTP
        """
        # initialize HOTP in slot
        self.set_mode(slot, yubico.yubikey_defs.MODE.OTP)
        cfg = self.key.init_config()
        if oath_tokenid:
            if not oath_tokenid.startswith('ubhe'):
                raise ValueError(
                    'Expected oath_tokenid with prefix "ubhe", but was %r' % (oath_tokenid)
                )
            cfg.mode_oath_hotp(
                secret=oath_secret,
                digits=otp_len,
                omp=0xe1,
                tt=0x63,
                mui=b'h:'+oath_tokenid[4:].encode('ascii'),
            )
            cfg.config_flag('OATH_FIXED_MODHEX2', True)
        else:
            cfg.mode_oath_hotp(secret=oath_secret, digits=otp_len)
        cfg.access_key(access_code.encode('ascii'))
        cfg.ticket_flag('APPEND_CR', False)
        cfg.ticket_flag('PROTECT_CFG2', True)
        cfg.extended_flag('SERIAL_API_VISIBLE', True)
        cfg.extended_flag('ALLOW_UPDATE', False)
        cfg.extended_flag('FAST_TRIG', True)
        self._write_cfg(cfg, slot)
        # end of init_hotp()

    def is_enabled(self, slot):
        """
        check wether slot is already enabled/configured
        """
        return slot in self.key.status().valid_configs()

    def _write_scancode(self, scancode, slot, access_code, current_access_code):
        # Check scancode argument
        if not scancode.startswith('h:') or scancode == 'h:' or \
           len(scancode) > 78 or  (len(scancode) % 2) != 0:
            raise ValueError('Invalid scancode %s' % repr(scancode))
        scancode = scancode[2:]
        cfg = self.key.init_config()
        cfg.enable_extended_scan_code_mode()
        if len(scancode) > 32:
            fixstr = scancode[:32]
            scancode = scancode[32:]
            cfg.fixed_string('h:' + fixstr)
        elif len(scancode) >= 1:
            fixstr = scancode
            scancode = ''
            cfg.fixed_string('h:' + fixstr)
        if len(scancode) > 12:
            uidstr = scancode[:12]
            scancode = scancode[12:]
            cfg.uid = unhexlify(uidstr)
        elif len(scancode) >= 1:
            uidstr = scancode
            scancode = ''
            cfg.uid = unhexlify(uidstr)
        if len(scancode) > 32:
            aeskeystr = scancode[:32]
            scancode = scancode[32:]
            cfg.aes_key('h:' + aeskeystr)
        elif len(scancode) >= 1:
            aeskeystr = scancode.ljust(32, '0')
            scancode = ''
            cfg.aes_key('h:' + aeskeystr)
        if self.is_enabled(slot) and current_access_code:
            cfg.unlock_key(current_access_code.encode('ascii'))
        cfg.extended_flag('SERIAL_API_VISIBLE', True)
        cfg.extended_flag('ALLOW_UPDATE', False)
        cfg.access_key(access_code.encode('ascii'))
        self._write_cfg(cfg, slot)
        # end of _write_scancode()

    def reset_slot(self, slot, current_access_code):
        """
        reset the given :slot:
        """
        self._write_scancode(
            YK_SCAN_CODE,
            slot,
            'h:000000000000',
            current_access_code,
        )
        time.sleep(1)
        cfg = self.key.init_config(zap=True)
        self._write_cfg(cfg, slot)
        # end of reset_slot()

    def clear(self, current_access_code):
        """
        clear/reset all slots in the device
        """
        for slot in self.enabled_slots():
            self.reset_slot(slot, current_access_code)
        # end of clear()

    def set_mode(self, slot, mode):
        """
        set operation mode of a slot
        """
        ykcfg = self.key.init_device_config(mode=mode)
        self._write_cfg(ykcfg, slot)
        # end of set_mode()

    def set_nfc(self, slot, current_access_code):
        """
        assign NFC to a slot
        """
        if self.key.model == 'YubiKey 4':
            # NDEF slot 2 unsupported in YubiKey 4
            return
        ykcfg_ndef = YubiKeyNEO_NDEF(data=b'', access_code=current_access_code)
        self._write_cfg(ykcfg_ndef, slot=slot)
        # end of set_nfc()

    def initialize(self, oath_secret, otp_len, oath_tokenid, access_code):
        """
        do the complete initialization of both slots
        """
        self.init_hotp(1, oath_secret, otp_len, oath_tokenid, access_code)
        # NFC tied to disabled 2nd slot
        self.set_nfc(2, access_code.encode('ascii'))
        # end of finalize()
